import json
import os
import pathlib

from fhir.resources.bundle import Bundle

from gen3_tracker.common import read_ndjson_file
from tests.integration import run
from click.testing import CliRunner
import pytest


# user made a mistake and changed the patient id,  "P1" to "P1-prime".  This will create a bundle with two entries, to delete the entries for P1 one for the patient and one for the ResearchSubject.
CHANGE_PATIENT = [
    "--debug add s3://s3-bucket/p1-object.txt --size 1 --modified 2024-05-05T07:26:29-0700 --md5 acbd18db4cc2f85cedef654fccc4a4d8 --patient P1",
    "--debug meta init",
    "--debug commit -am \"initial commit\"",
    "--debug add s3://s3-bucket/p1-object.txt --size 1 --modified 2024-05-05T07:26:29-0700 --md5 acbd18db4cc2f85cedef654fccc4a4d8 --patient P1-prime",
    "--debug meta init",
    "--debug commit -am \"prime commit\"",
]

# user made a mistake and added the wrong file
CHANGE_FILE = [
    "--debug add s3://s3-bucket/p1-object-mistake.txt --size 1 --modified 2024-05-05T07:26:29-0700 --md5 acbd18db4cc2f85cedef654fccc4a4d8 --patient P1",
    "--debug meta init",
    "--debug commit -am \"initial commit\"",
    "--debug rm s3://s3-bucket/p1-object-mistake.txt",
    "--debug add s3://s3-bucket/p1-object-correct.txt --size 1 --modified 2024-05-05T07:26:29-0700 --md5 acbd18db4cc2f85cedef654fccc4a4d8 --patient P1",
    "--debug meta init",
    "--debug commit -am \"prime commit\"",
]


def test_change_patient(runner: CliRunner, project_id, tmpdir) -> None:
    """Ensure bundle gets created when `meta init` orphans records."""
    # change to the temporary directory
    assert tmpdir.chdir()
    print(pathlib.Path.cwd())

    print(project_id)

    run(runner, ["--debug", "init", project_id, "--approve", "--no-server"],
        expected_files=[".g3t", ".git"])

    for _ in CHANGE_PATIENT:
        run(runner, _.split())

    for _ in read_ndjson_file(pathlib.Path("META/Bundle.ndjson")):
        print(_)
        bundle = Bundle(**_)
        break  # only one bundle

    assert len(bundle.entry) == 2, "Only two entries are expected."

    methods = [_.request.method for _ in bundle.entry]
    assert all([_ == "DELETE" for _ in methods]), "Only DELETE method is expected."

    urls = [_.request.url for _ in bundle.entry]
    assert any([_.startswith('Patient') for _ in urls]), "Expected to delete a Patient."
    assert any([_.startswith('ResearchSubject') for _ in urls]), "Expected to delete a ResearchSubject."


def test_change_file(runner: CliRunner, project_id, tmpdir) -> None:
    """Ensure bundle gets created when `meta init` orphans records.."""
    # change to the temporary directory
    assert tmpdir.chdir()
    print(pathlib.Path.cwd())

    print(project_id)

    run(runner, ["--debug", "init", project_id, "--approve", "--no-server"],
        expected_files=[".g3t", ".git"])

    for _ in CHANGE_FILE:
        run(runner, _.split())

    for _ in read_ndjson_file(pathlib.Path("META/Bundle.ndjson")):
        print(_)
        bundle = Bundle(**_)
        break  # only one bundle

    assert len(bundle.entry) == 1, "Only one entry is expected."

    methods = [_.request.method for _ in bundle.entry]
    assert all([_ == "DELETE" for _ in methods]), "Only DELETE method is expected."

    urls = [_.request.url for _ in bundle.entry]
    assert any([_.startswith('DocumentReference') for _ in urls]), "Expected to delete a DocumentReference."



from typing import List, Any, Union
from collections import defaultdict

from gen3.auth import Gen3Auth
from pydantic import BaseModel

from gen3_tracker.config import ensure_auth, Config


class ProjectSummary(BaseModel):
    """Summary of a project."""
    exists: bool = False
    """Project exists in sheepdog flag"""
    permissions: list[dict[str, Any]] = []


class ProjectSummaries(BaseModel):
    """Summary of projects, including messages."""
    endpoint: str
    """The commons url"""
    incomplete: Union[dict[str, ProjectSummary], list[str]] = {}
    """List of projects that require creation in sheepdog."""
    complete: Union[dict[str, ProjectSummary], list[str]] = {}
    """List of projects that exist in sheepdog."""
    messages: List[str] = []
    """List of messages"""


def get_user(config: Config = None, auth: Gen3Auth = None) -> dict:
    """Fetch information about the user."""
    if not auth:
        assert config
        auth = ensure_auth(config=config)
    return auth.curl('/user/user').json()


def recursive_defaultdict():
    return defaultdict(recursive_defaultdict)


def get_projects(auth, submission) -> dict:
    """Return a dict of programs, projects and their existence flag."""

    # get the list of programs in sheepdog
    response = submission.get_programs()
    assert 'links' in response, f'submission.get_program returned unexpected response: {response}'
    program_links = response['links']
    programs = [_.split('/')[-1] for _ in program_links]
    project_links = []
    sheepdog_projects = recursive_defaultdict()
    # add projects to it
    for program in programs:
        # print(program)
        project_links.extend(submission.get_projects(program)['links'])
        sheepdog_projects[program]['in_sheepdog'] = True
        sheepdog_projects[program]['projects'] = {}
    for _ in project_links:
        program, project = _.replace('/v0/submission/', '').split('/')
        sheepdog_projects[program]['projects'][project] = True

    # get the list of programs and projects from arborist, will be different from sheepdog
    user = get_user(auth=auth)
    arborist_projects = recursive_defaultdict()
    for _ in user['authz'].keys():
        if not all([_.startswith('/programs'), 'projects/' in _]):
            continue

        permissions = user['authz'][_]

        _ = _.replace('/programs/', '')
        _ = _.split('/')
        _program = _[0]
        _project = _[-1]
        arborist_projects[_program][_project]['permissions'] = permissions
        arborist_projects[_program][_project]['exists'] = False
        if _program in sheepdog_projects and _project in sheepdog_projects[_program]['projects']:
            arborist_projects[_program][_project]['exists'] = True

    return arborist_projects



from datetime import datetime, date
from urllib.parse import urlparse

import pytest
from fhir.resources.coding import Coding
from fhir.resources.domainresource import DomainResource
from fhir.resources.identifier import Identifier
from fhir.resources.observation import Observation
from fhir.resources.patient import Patient
from fhir.resources.reference import Reference
from fhir.resources.specimen import Specimen


# Implement "flatten" by monkey patching the DomainResource class.
# Specialize the flatten method for Observation.

#            __,__
#   .--.  .-"     "-.  .--.
#  / .. \/  .-. .-.  \/ .. \
# | |  '|  /   Y   \  |'  | |
# | \   \  \ 0 | 0 /  /   / |
#  \ '- ,\.-"`` ``"-./, -' /
#   `'-' /_   ^ ^   _\ '-'`
#       |  \._   _./  |
#       \   \ `~` /   /
#        '._ '-=-' _.'
#           '~---~'

# test data ------------------------------------------------------------
# The following fixtures provide test data for the tests below.

@pytest.fixture
def patient_dict() -> dict:
    # TODO - read the patient example from a file
    patient_dict = {"resourceType": "Patient", "id": "3", "meta": {"lastUpdated": "2012-05-29T23:45:32Z"},
                    "text": {"status": "generated",
                             "div": "\u003cdiv xmlns\u003d\"http://www.w3.org/1999/xhtml\"\u003eKidd, Kari. SSN:\n            444555555\u003c/div\u003e"},
                    "identifier": [{"type": {
                        "coding": [{"system": "http://terminology.hl7.org/CodeSystem/v2-0203", "code": "SS"}]},
                                    "system": "http://hl7.org/fhir/sid/us-ssn", "value": "444555555"}], "active": True,
                    "name": [{"use": "official", "family": "Kidd", "given": ["Kari"]}],
                    "telecom": [{"system": "phone", "value": "555-555-2005", "use": "work"}], "gender": "female",
                    "address": [{"use": "home", "line": ["2222 Home Street"]}],
                    "managingOrganization": {"reference": "Organization/hl7"}}
    yield patient_dict


@pytest.fixture
def specimen_dict():
    return {"resourceType": "Specimen", "id": "denovo-3", "text": {"status": "generated",
                                                                   "div": "\u003cdiv xmlns\u003d\"http://www.w3.org/1999/xhtml\"\u003e\u003cp\u003e\u003cb\u003eGenerated Narrative\u003c/b\u003e\u003c/p\u003e\u003cdiv style\u003d\"display: inline-block; background-color: #d9e0e7; padding: 6px; margin: 4px; border: 1px solid #8da1b4; border-radius: 5px; line-height: 60%\"\u003e\u003cp style\u003d\"margin-bottom: 0px\"\u003eResource \u0026quot;denovo-3\u0026quot; \u003c/p\u003e\u003c/div\u003e\u003cp\u003e\u003cb\u003eidentifier\u003c/b\u003e: id: 3\u003c/p\u003e\u003cp\u003e\u003cb\u003estatus\u003c/b\u003e: available\u003c/p\u003e\u003cp\u003e\u003cb\u003etype\u003c/b\u003e: Venous blood specimen \u003cspan style\u003d\"background: LightGoldenRodYellow; margin: 4px; border: 1px solid khaki\"\u003e (\u003ca href\u003d\"https://browser.ihtsdotools.org/\"\u003eSNOMED CT\u003c/a\u003e#122555007)\u003c/span\u003e\u003c/p\u003e\u003cp\u003e\u003cb\u003esubject\u003c/b\u003e: \u003ca href\u003d\"Patient-denovoFather.html\"\u003ePatient/denovoFather: John Doe\u003c/a\u003e \u0026quot; DOE\u0026quot;\u003c/p\u003e\u003cp\u003e\u003cb\u003ereceivedTime\u003c/b\u003e: 2021-01-01 01:01:01+0000\u003c/p\u003e\u003cp\u003e\u003cb\u003erequest\u003c/b\u003e: \u003ca href\u003d\"ServiceRequest-genomicServiceRequest.html\"\u003eServiceRequest/genomicServiceRequest\u003c/a\u003e\u003c/p\u003e\u003ch3\u003eCollections\u003c/h3\u003e\u003ctable class\u003d\"grid\"\u003e\u003ctr\u003e\u003ctd\u003e-\u003c/td\u003e\u003ctd\u003e\u003cb\u003eCollector\u003c/b\u003e\u003c/td\u003e\u003ctd\u003e\u003cb\u003eCollected[x]\u003c/b\u003e\u003c/td\u003e\u003ctd\u003e\u003cb\u003eQuantity\u003c/b\u003e\u003c/td\u003e\u003ctd\u003e\u003cb\u003eMethod\u003c/b\u003e\u003c/td\u003e\u003c/tr\u003e\u003ctr\u003e\u003ctd\u003e*\u003c/td\u003e\u003ctd\u003e\u003ca href\u003d\"Practitioner-practitioner01.html\"\u003ePractitioner/practitioner01\u003c/a\u003e \u0026quot; DOEL\u0026quot;\u003c/td\u003e\u003ctd\u003e2021-01-01 01:01:00+0000\u003c/td\u003e\u003ctd\u003e1 mL\u003c/td\u003e\u003ctd\u003eLine, Venous \u003cspan style\u003d\"background: LightGoldenRodYellow; margin: 4px; border: 1px solid khaki\"\u003e (\u003ca href\u003d\"http://terminology.hl7.org/3.1.0/CodeSystem-v2-0488.html\"\u003especimenCollectionMethod\u003c/a\u003e#LNV)\u003c/span\u003e\u003c/td\u003e\u003c/tr\u003e\u003c/table\u003e\u003c/div\u003e"},
            "identifier": [{"system": "http://www.somesystemabc.net/identifiers/specimens", "value": "3"}],
            "status": "available", "type": {
            "coding": [{"system": "http://snomed.info/sct", "code": "122555007", "display": "Venous blood specimen"}]},
            "subject": {"reference": "Patient/denovoFather", "display": "John Doe"},
            "receivedTime": "2021-01-01T01:01:01Z", "request": [{"reference": "ServiceRequest/genomicServiceRequest"}],
            "collection": {"collector": {"reference": "Practitioner/practitioner01"},
                           "collectedDateTime": "2021-01-01T01:01:00Z", "quantity": {"value": 1, "unit": "mL"},
                           "method": {
                               "coding": [{"system": "http://terminology.hl7.org/CodeSystem/v2-0488", "code": "LNV"}]}}}


@pytest.fixture
def observation_eye_color_dict():
    return {"resourceType": "Observation", "id": "eye-color", "text": {"status": "generated",
                                                                       "div": "\u003cdiv xmlns\u003d\"http://www.w3.org/1999/xhtml\"\u003e\u003cp\u003e\u003cb\u003eGenerated Narrative: Observation\u003c/b\u003e\u003ca name\u003d\"eye-color\"\u003e \u003c/a\u003e\u003ca name\u003d\"hceye-color\"\u003e \u003c/a\u003e\u003c/p\u003e\u003cdiv style\u003d\"display: inline-block; background-color: #d9e0e7; padding: 6px; margin: 4px; border: 1px solid #8da1b4; border-radius: 5px; line-height: 60%\"\u003e\u003cp style\u003d\"margin-bottom: 0px\"\u003eResource Observation \u0026quot;eye-color\u0026quot; \u003c/p\u003e\u003c/div\u003e\u003cp\u003e\u003cb\u003estatus\u003c/b\u003e: final\u003c/p\u003e\u003cp\u003e\u003cb\u003ecode\u003c/b\u003e: eye color \u003cspan style\u003d\"background: LightGoldenRodYellow; margin: 4px; border: 1px solid khaki\"\u003e ()\u003c/span\u003e\u003c/p\u003e\u003cp\u003e\u003cb\u003esubject\u003c/b\u003e: \u003ca href\u003d\"patient-example.html\"\u003ePatient/example\u003c/a\u003e \u0026quot;Peter CHALMERS\u0026quot;\u003c/p\u003e\u003cp\u003e\u003cb\u003eeffective\u003c/b\u003e: 2016-05-18\u003c/p\u003e\u003cp\u003e\u003cb\u003evalue\u003c/b\u003e: blue\u003c/p\u003e\u003c/div\u003e"},
            "status": "final", "code": {"text": "eye color"}, "subject": {"reference": "Patient/example"},
            "effectiveDateTime": "2016-05-18", "valueString": "blue"}


@pytest.fixture
def observation_bmi_dict():
    return {"resourceType": "Observation", "id": "bmi-using-related", "text": {"status": "generated",
                                                                               "div": "\u003cdiv xmlns\u003d\"http://www.w3.org/1999/xhtml\"\u003e\u003cp\u003e\u003cb\u003eGenerated Narrative: Observation\u003c/b\u003e\u003ca name\u003d\"bmi-using-related\"\u003e \u003c/a\u003e\u003ca name\u003d\"hcbmi-using-related\"\u003e \u003c/a\u003e\u003c/p\u003e\u003cdiv style\u003d\"display: inline-block; background-color: #d9e0e7; padding: 6px; margin: 4px; border: 1px solid #8da1b4; border-radius: 5px; line-height: 60%\"\u003e\u003cp style\u003d\"margin-bottom: 0px\"\u003eResource Observation \u0026quot;bmi-using-related\u0026quot; \u003c/p\u003e\u003c/div\u003e\u003cp\u003e\u003cb\u003estatus\u003c/b\u003e: \u003cspan title\u003d\"  \u003d\u003d\u003d\u003d\u003d\u003d\u003d\u003d need to fix vitals to removed fixed value \u0027has-member\u0027 \u003d\u003d\u003d\u003d\u003d\u003d\u003d\u003d\u003d\u003d\n\t\u0026lt;meta\u0026gt;\n\t\t\u0026lt;profile value\u003d\u0026quot;http://hl7.org/fhir/StructureDefinition/vitalsigns\u0026quot;/\u0026gt;\n\t\u0026lt;/meta\u0026gt;\n     \"\u003efinal\u003c/span\u003e\u003c/p\u003e\u003cp\u003e\u003cb\u003ecategory\u003c/b\u003e: Vital Signs \u003cspan style\u003d\"background: LightGoldenRodYellow; margin: 4px; border: 1px solid khaki\"\u003e (\u003ca href\u003d\"http://terminology.hl7.org/5.5.0/CodeSystem-observation-category.html\"\u003eObservation Category Codes\u003c/a\u003e#vital-signs)\u003c/span\u003e\u003c/p\u003e\u003cp\u003e\u003cb\u003ecode\u003c/b\u003e: BMI \u003cspan style\u003d\"background: LightGoldenRodYellow; margin: 4px; border: 1px solid khaki\"\u003e (\u003ca href\u003d\"https://loinc.org/\"\u003eLOINC\u003c/a\u003e#39156-5 \u0026quot;Body mass index (BMI) [Ratio]\u0026quot;)\u003c/span\u003e\u003c/p\u003e\u003cp\u003e\u003cb\u003esubject\u003c/b\u003e: \u003ca href\u003d\"patient-example.html\"\u003ePatient/example\u003c/a\u003e \u0026quot;Peter CHALMERS\u0026quot;\u003c/p\u003e\u003cp\u003e\u003cb\u003eeffective\u003c/b\u003e: 1999-07-02\u003c/p\u003e\u003cp\u003e\u003cb\u003evalue\u003c/b\u003e: 16.2 kg/m2\u003cspan style\u003d\"background: LightGoldenRodYellow\"\u003e (Details: UCUM code kg/m2 \u003d \u0027kg/m2\u0027)\u003c/span\u003e\u003c/p\u003e\u003cp\u003e\u003cb\u003ederivedFrom\u003c/b\u003e: \u003c/p\u003e\u003cul\u003e\u003cli\u003e\u003ca href\u003d\"broken-link.html\"\u003eObservation/bodyheight: Body Height\u003c/a\u003e\u003c/li\u003e\u003cli\u003e\u003ca href\u003d\"observation-example.html\"\u003eObservation/example: Body Weight\u003c/a\u003e\u003c/li\u003e\u003c/ul\u003e\u003c/div\u003e"},
            "status": "final", "category": [{"coding": [
            {"system": "http://terminology.hl7.org/CodeSystem/observation-category", "code": "vital-signs",
             "display": "Vital Signs"}], "text": "Vital Signs"}], "code": {
            "coding": [{"system": "http://loinc.org", "code": "39156-5", "display": "Body mass index (BMI) [Ratio]"}],
            "text": "BMI"}, "subject": {"reference": "Patient/example"}, "effectiveDateTime": "1999-07-02",
            "valueQuantity": {"value": 16.2, "unit": "kg/m2", "system": "http://unitsofmeasure.org", "code": "kg/m2"},
            "derivedFrom": [{"reference": "Observation/bodyheight", "display": "Body Height"},
                            {"reference": "Observation/example", "display": "Body Weight"}]}


# flatteners ------------------------------------------------------------
# The following functions are used to flatten the FHIR resources.

def flatten_simple(self: DomainResource):
    """Convert the DomainResource instance to just an id."""
    return self.id


def _isodate(v):
    """If a value is a datetime object, return it as an ISO 8601 string."""
    if isinstance(v, datetime) or isinstance(v, date):
        return v.isoformat()
    return v


def flatten_scalars(self: DomainResource) -> dict:
    """Convert the DomainResource instance to a dictionary."""
    _ = {k: _isodate(v) for k, v in self.dict().items() if not isinstance(v, (list, dict))}
    return _


def flatten_references(self: DomainResource) -> dict:
    """Convert the DomainResource instance to a dictionary."""
    fields = [_ for _ in self.__fields__.keys() if not _.endswith('__ext')]
    _ = {}
    # if any top level field in this resource is a Reference, use the Reference.reference https://build.fhir.org/references-definitions.html#Reference.reference
    for k in fields:
        v = getattr(self, k)
        if isinstance(v, Reference):
            v: Reference = v
            _[k] = v.reference
    return _


def flatten_identifier(self: Identifier) -> dict:
    """Convert the Identifier instance to a key value, use a simplified system as key."""
    parsed_url = urlparse(self.system)
    path_parts = parsed_url.path.split('/')  # e.g. "http://hl7.org/fhir/sid/us-ssn" -> us-ssn
    key = path_parts[-1] if path_parts else 'identifier'
    return {key: self.value}


def flatten_coding(self: Coding) -> dict:
    """Convert the DomainResource instance to a dictionary."""
    return {'display': self.display}


def flatten_scalars_and_references(self: DomainResource) -> dict:
    """Convert the DomainResource instance to a dictionary."""
    _ = flatten_scalars(self)
    _.update(flatten_references(self))
    return _


def flatten_scalars_references_identifiers(self: DomainResource) -> dict:
    """Convert the DomainResource instance to a dictionary."""
    _ = flatten_scalars_and_references(self)
    # for now, just flatten all identifiers
    if self.identifier:
        for identifier in self.identifier:
            _.update(flatten_identifier(identifier))
    return _


def flatten_observation(self: Observation) -> dict:
    """Convert the DomainResource instance to a dictionary."""
    _ = flatten_scalars_references_identifiers(self)
    # normalize all the valueXXXXX to 'value'
    if self.valueQuantity:
        _['value'] = f"{self.valueQuantity.value} {self.valueQuantity.unit}"
    elif self.valueString:
        _['value'] = self.valueString
        del _['valueString']
    elif self.valueCodeableConcept:
        _['value'] = self.valueCodeableConcept.text
    # there are many other value types, but we'll ignore them for now
    # see https://build.fhir.org/observation-definitions.html#Observation.value_x_
    # 	Quantity|CodeableConcept|string|boolean|integer|Range|Ratio|SampledData|time|dateTime|Period|Attachment|Reference(MolecularSequence)
    # [for _ in self.__fields__.keys() if _.startswith('value')]
    return _


# patchers ------------------------------------------------------------
# The following fixtures are used to patch the DomainResource class to add the desired method.

@pytest.fixture
def patched_domain_resource_simple() -> bool:
    """Patch the DomainResource class to add a flatten method."""
    # Add the flatten method to the DomainResource class
    DomainResource.flatten = flatten_simple
    yield True
    # Remove the flatten method from the DomainResource class
    del DomainResource.flatten


@pytest.fixture
def patched_scalars() -> bool:
    """Patch the DomainResource class to add a flatten method."""
    # Add the flatten method to the DomainResource class
    DomainResource.flatten = flatten_scalars
    yield True
    # Remove the flatten method from the DomainResource class
    del DomainResource.flatten


@pytest.fixture
def patched_scalars_and_references() -> bool:
    """Patch the DomainResource class to add a flatten method."""
    # Add the flatten method to the DomainResource class
    DomainResource.flatten = flatten_scalars_and_references
    yield True
    # Remove the flatten method from the DomainResource class
    del DomainResource.flatten


@pytest.fixture
def patched_scalars_references_identifiers() -> bool:
    """Patch the DomainResource class to add a flatten method."""
    # Add the flatten method to the DomainResource class
    DomainResource.flatten = flatten_scalars_references_identifiers
    yield True
    # Remove the flatten method from the DomainResource class
    del DomainResource.flatten


@pytest.fixture
def patched_scalars_references_identifiers_observation() -> bool:
    """Patch the DomainResource class to add a flatten method."""
    # Add the flatten method to the DomainResource class e.g. all resources get flattened this way
    DomainResource.flatten = flatten_scalars_references_identifiers
    # except for Observation, which has a different flatten method
    Observation.flatten = flatten_observation
    yield True
    # Remove the flatten method from the DomainResource class
    del DomainResource.flatten
    del Observation.flatten


# tests ------------------------------------------------------------

def test_patient_without_flatten(patient_dict: dict):
    """This patient object should NOT have a 'flatten' method."""
    # without path dependency, just have a plain patient object with no flatten method
    patient = Patient.parse_obj(patient_dict)
    assert not hasattr(patient, 'flatten'), "Patient object should not have a 'flatten' method"


def test_patient_with_simple(patched_domain_resource_simple: bool, patient_dict: dict):
    """This patient object should have a 'flatten' method."""
    patient = Patient.parse_obj(patient_dict)
    assert hasattr(patient, 'flatten'), "Patient object does not have a 'flatten' method"
    assert patient.flatten() == patient.id, f"Patient.flatten() should return {patient.id}"


def test_patient_with_scalars(patched_scalars: bool, patient_dict: dict):
    """This patient object should have a 'flatten' method that returns a dict of scalar values."""
    patient = Patient.parse_obj(patient_dict)
    assert hasattr(patient, 'flatten'), "Patient object does not have a 'flatten' method"
    assert patient.flatten() == {'active': True, 'gender': 'female', 'id': '3', 'resourceType': 'Patient'}, "Patient.flatten() should return a dict of all scalar values"


def test_patient_with_scalars_and_references(patched_scalars_and_references: bool, patient_dict: dict):
    """This patient object should have a 'flatten' method that returns a dict of scalar values and references."""
    patient = Patient.parse_obj(patient_dict)
    assert hasattr(patient, 'flatten'), "Patient object does not have a 'flatten' method"
    assert patient.flatten() == {'active': True, 'gender': 'female', 'id': '3', 'managingOrganization': 'Organization/hl7', 'resourceType': 'Patient'}, "Patient.flatten() should return a dict of all scalar values and references"


def test_patient_with_scalars_references_identifiers(patched_scalars_references_identifiers: bool, patient_dict: dict):
    """This patient object should have a 'flatten' method that returns a dict of scalar values and references."""
    patient = Patient.parse_obj(patient_dict)
    assert hasattr(patient, 'flatten'), "Patient object does not have a 'flatten' method"
    assert patient.flatten() == {'active': True, 'gender': 'female', 'id': '3', 'managingOrganization': 'Organization/hl7', 'resourceType': 'Patient', 'us-ssn': '444555555'}, "Patient.flatten() should return a dict of all scalar values and references"


def test_specimen_with_scalars_references_identifiers(patched_scalars_references_identifiers: bool, specimen_dict: dict):
    """This patient object should have a 'flatten' method that returns a dict of scalar values and references."""
    specimen = Specimen.parse_obj(specimen_dict)
    assert hasattr(specimen, 'flatten'), "Specimen object does not have a 'flatten' method"
    assert specimen.flatten() == {'resourceType': 'Specimen', 'id': 'denovo-3', 'status': 'available',
                                  'receivedTime': '2021-01-01T01:01:01+00:00',
                                  'subject': 'Patient/denovoFather', 'specimens': '3'}


def test_eye_color_observation(patched_scalars_references_identifiers_observation: bool, observation_eye_color_dict: dict):
    """This patient object should have a 'flatten' method that returns a dict of scalar values and references."""
    observation = Observation.parse_obj(observation_eye_color_dict)
    assert hasattr(observation, 'flatten'), "Observation object does not have a 'flatten' method"
    assert observation.flatten() == {'resourceType': 'Observation', 'id': 'eye-color', 'status': 'final',
                                     'effectiveDateTime': '2016-05-18', 'value': 'blue',
                                     'subject': 'Patient/example'}


def test_bmi_observation(patched_scalars_references_identifiers_observation: bool, observation_bmi_dict: dict):
    """This patient object should have a 'flatten' method that returns a dict of scalar values and references."""
    observation = Observation.parse_obj(observation_bmi_dict)
    assert hasattr(observation, 'flatten'), "Observation object does not have a 'flatten' method"
    assert observation.flatten() == {'effectiveDateTime': '1999-07-02',
                                     'id': 'bmi-using-related',
                                     'resourceType': 'Observation',
                                     'status': 'final',
                                     'subject': 'Patient/example',
                                     'value': '16.2 kg/m2'}



import json
import pathlib
import sqlite3
import uuid
from functools import lru_cache
from typing import Generator, Optional, List, Tuple

import inflection
import ndjson
import numpy as np
import pandas as pd
from tqdm import tqdm
from copy import deepcopy


class LocalFHIRDatabase:
    def __init__(self, db_name):  # , db_name=pathlib.Path('.g3t') / 'local.db'):
        self.db_name = db_name
        self.connection = None
        self.cursor = None
        self.table_created = {}  # Flag to track if the table has been created

    def connect(self) -> sqlite3.Cursor:
        self.connection = sqlite3.connect(self.db_name)
        self.cursor = self.connection.cursor()

    def disconnect(self):
        if self.connection:
            self.connection.commit()
            self.connection.close()

    def create_table(
            self,
            name='resources',
            ddl='''
                    CREATE TABLE __NAME__ (
                        key TEXT PRIMARY KEY,
                        resource_type TEXT,
                        resource JSON
                    )
                '''):
        self.connect()
        # Check if the table exists before creating it
        self.cursor.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name='{name}'")
        table_exists = self.cursor.fetchone()

        if not table_exists:
            ddl = ddl.replace('__NAME__', name)
            self.cursor.execute(ddl)
            self.table_created[name] = True

    def count(self, table_name='resources'):
        self.connect()
        self.cursor.execute(f"SELECT COUNT(*) FROM {table_name}")
        count = self.cursor.fetchone()[0]
        return count

    def insert_data(self, id_, resource_type, resource, table_name='resources'):
        """Insert data into the database."""
        if table_name not in self.table_created:
            self.create_table(table_name)  # Lazily create the table if not already created

        composite_key = f"{resource_type}/{id_}"
        self.cursor.execute(f'''
            INSERT INTO {table_name} (key, resource_type, resource)
            VALUES (?, ?, ?)
        ''', (composite_key, resource_type, json.dumps(resource)))
        # print(f"Inserted {composite_key} into the database")

    def insert_data_from_dict(self, resource, table_name='resources'):
        """Insert data into the database from a dictionary."""
        if 'id' not in resource or ('resource_type' not in resource and 'resourceType' not in resource):
            raise ValueError(f"Resource dictionary must contain 'id' and 'resource_type' keys {resource}")
        self.insert_data(
            resource['id'],
            resource.get('resource_type', resource.get('resourceType')),
            resource,
            table_name
        )

    def bulk_insert_data(self, resources, table_name='resources') -> int:
        """Bulk insert data into the database."""

        if table_name not in self.table_created:
            self.create_table(table_name)  # Lazily create the table if not already created

        def _prepare(resource):
            resource_type = resource.get('resource_type', resource.get('resourceType'))
            id_ = resource['id']
            composite_key = f"{resource_type}/{id_}"
            return (
                composite_key,
                resource_type,
                json.dumps(resource)
            )

        def _iterate(_resources):
            for _ in _resources:
                yield _prepare(_)

        try:
            self.connect()
            sql = f'''
                INSERT INTO {table_name} (key, resource_type, resource)
                VALUES (?, ?, ?)
            '''
            new_cursor = self.cursor.executemany(sql, _iterate(_resources=resources))

        except sqlite3.IntegrityError as e:
            for resource in resources:
                prepared_resource = _prepare(resource)
                try:
                    self.cursor.execute(sql, prepared_resource)
                except sqlite3.IntegrityError:
                    print(f"Error inserting resource: {prepared_resource}")
                    print(f"Exception: {e}")
            raise

        finally:
            self.connection.commit()
            # self.disconnect()

        return new_cursor.rowcount

    def load_from_ndjson_file(self, file_path, table_name='resources'):
        """Load the NDJSON file into the database."""

        if table_name not in self.table_created:
            self.create_table(table_name)  # Lazily create the table if not already created

        with open(file_path, 'r') as file:
            reader = ndjson.reader(file)
            self.bulk_insert_data(reader)

    def load_ndjson_from_dir(self, path: str = 'META', pattern: str = '*.ndjson'):
        """Load all the NDJSON files in the directory into the database."""
        for file_path in pathlib.Path(path).glob(pattern):
            self.load_from_ndjson_file(file_path)

    @lru_cache(maxsize=None)
    def patient_everything(self, patient_id) -> Generator[dict, None, None]:
        """Return all the resources for a patient."""
        cursor = self.connection.cursor()
        cursor.execute("SELECT * FROM resources WHERE key = ?",
                       (f"Patient/{patient_id}",))

        for _ in cursor.fetchall():
            key, resource_type, resource = _
            yield json.loads(resource)

    @lru_cache(maxsize=None)
    def patient(self, patient_id) -> dict:
        """Return the patient resource."""
        cursor = self.connection.cursor()
        cursor.execute("SELECT * FROM resources WHERE key = ?", (f"Patient/{patient_id}",))
        _ = cursor.fetchone()
        if _ is None:
            print(f"Patient {patient_id} not found")
            return None
        key, resource_type, resource = _
        resource = json.loads(resource)

        resource = self.simplify_extensions(resource)

        return resource

    @staticmethod
    def simplify_extensions(resource: dict) -> dict:
        """Extract extension values, derive key from extension url"""
        for _ in resource.get('extension', []):
            value_normalized, value_source = normalize_value(_)
            extension_key = _['url'].split('/')[-1]
            extension_key = inflection.underscore(extension_key)
            resource[extension_key] = value_normalized
            assert value_normalized, f"extension: {extension_key} = {value_normalized}"
        if 'extension' in resource:
            del resource['extension']
        return resource

    @lru_cache(maxsize=None)
    def flattened_procedure(self, procedure_key) -> dict:
        """Return the procedure with everything resolved."""
        cursor = self.connection.cursor()
        cursor.execute("SELECT * FROM resources WHERE key = ?", (procedure_key,))
        key, resource_type, resource = cursor.fetchone()
        procedure = json.loads(resource)

        # simplify the identifier
        procedure['identifier'] = procedure['identifier'][0]['value']
        # simplify the code
        procedure['code'] = procedure['code']['coding'][0]['display']
        # simplify the reason
        procedure['reason'] = procedure['reason'][0]['reference']['reference']
        # simplify the occurrenceAge
        procedure['occurrenceAge'] = procedure['occurrenceAge']['value']
        # simplify the subject
        subject = procedure['subject']['reference']
        procedure['subject'] = subject
        return procedure

    @lru_cache(maxsize=None)
    def flattened_specimen(self, specimen_key) -> dict:
        """Return the procedure with everything resolved."""
        cursor = self.connection.cursor()
        cursor.execute("SELECT * FROM resources WHERE key = ?", (specimen_key,))
        key, resource_type, resource = cursor.fetchone()
        specimen = json.loads(resource)

        # simplify the identifier
        specimen['identifier'] = specimen['identifier'][0]['value']

        for coding_normalized, coding_source in normalize_coding(specimen['collection']):
            specimen[f"collection_{coding_source}"] = coding_normalized
        del specimen['collection']
        for processing in specimen.get('processing', []):
            for coding_normalized, coding_source in normalize_coding(processing):
                specimen[f"processing_{coding_source}"] = coding_normalized
            break # TODO - only first one
        del specimen['processing']

        return specimen

    @lru_cache(maxsize=None)
    def flattened_condition(self, condition_key) -> dict:
        """Return the procedure with everything resolved."""
        cursor = self.connection.cursor()
        cursor.execute("SELECT * FROM resources WHERE key = ?", (condition_key,))
        key, resource_type, resource = cursor.fetchone()
        condition = json.loads(resource)

        # simplify the identifier
        condition['identifier'] = condition['identifier'][0]['value']
        # simplify the code
        condition['code'] = condition['code']['coding'][0]['display']
        for coding_normalized, coding_source in normalize_coding(condition):
            condition[coding_source] = coding_normalized
        # simplify the onsetAge
        condition['onsetAge'] = condition['onsetAge']['value']
        return condition

    def flattened_procedures(self) -> Generator[dict, None, None]:
        """Return all the procedures with everything resolved"""
        loaded_db = self
        connection = sqlite3.connect(loaded_db.db_name)
        cursor = connection.cursor()
        cursor.execute("SELECT * FROM resources where resource_type = ?", ("Procedure",))

        for _ in cursor.fetchall():
            key, resource_type, procedure = _
            procedure = json.loads(procedure)
            # simplify the identifier
            procedure['identifier'] = procedure['identifier'][0]['value']
            # simplify the code
            procedure['code'] = procedure['code']['coding'][0]['display']
            # simplify the reason
            procedure['reason'] = procedure['reason'][0]['reference']['reference']
            # simplify the occurrenceAge
            procedure['occurrenceAge'] = procedure['occurrenceAge']['value']
            # simplify the subject
            subject = procedure['subject']['reference']
            procedure['subject'] = subject

            if subject.startswith('Patient/'):
                _, patient_id = subject.split('/')
                resources = [_ for _ in loaded_db.patient_everything(patient_id)]
                resources.append(loaded_db.patient(patient_id))
                for resource in resources:

                    if resource['resourceType'] == 'Patient':
                        procedure['patient'] = resource['identifier'][0]['value']
                        continue

                    if resource['resourceType'] == 'Condition' and f"Condition/{resource['id']}" == procedure['reason']:
                        procedure['reason'] = resource['code']['text']
                        continue

                    if resource['resourceType'] == 'Observation':
                        # must be focus
                        if f"Procedure/{procedure['id']}" not in [_['reference'] for _ in resource['focus']]:
                            continue

                        # TODO - pick first coding, h2 allow user to specify preferred coding
                        code = resource['code']['coding'][0]['code']

                        if 'valueQuantity' in resource:
                            value = resource['valueQuantity']['value']
                        elif 'valueCodeableConcept' in resource:
                            value = resource['valueCodeableConcept']['text']
                        elif 'valueInteger' in resource:
                            value = resource['valueInteger']
                        elif 'valueString' in resource:
                            value = resource['valueString']
                        else:
                            value = None

                        assert value is not None, f"no value for {resource['id']}"
                        procedure[code] = value

                        continue

                    # skip these
                    if resource['resourceType'] in ['Specimen', 'Procedure', 'ResearchSubject']:
                        continue

                    # default, add entire resource as an item of the list
                    resource_type = inflection.underscore(resource['resourceType'])
                    if resource_type not in procedure:
                        procedure[resource_type] = []
                    procedure[resource_type].append(resource)

            yield procedure

        connection.close()

    def flattened_observations(self) -> Generator[dict, None, None]:
        loaded_db = self
        connection = sqlite3.connect(loaded_db.db_name)
        cursor = connection.cursor()
        cursor.execute("SELECT * FROM resources where resource_type = ?", ("Observation",))
        for _ in tqdm(cursor.fetchall()):
            key, resource_type, observation_str = _
            observation = json.loads(observation_str)

            # simplify the subject
            subject = observation['subject']['reference']
            observation['subject'] = subject

            # simplify the identifier
            observation['identifier'] = observation.get('identifier', [{'value': None}])[0]['value']

            # simplify the value
            value_normalized, value_source = normalize_value(observation)

            observation['value_normalized'] = value_normalized

            if value_normalized:
                value_numeric = observation['value_normalized'].split(' ')[0]
                if is_number(value_numeric):
                    observation['value_numeric'] = float(value_numeric)
                else:
                    observation['value_numeric'] = None
            else:
                observation['value_numeric'] = None

            if value_source:
                del observation[value_source]

            ignored_fields = ['meta', 'encounter']
            for _ in ignored_fields:
                if _ in observation:
                    del observation[_]

            observation = self.simplify_extensions(observation)

            if subject.startswith('Patient/'):
                _, patient_id = subject.split('/')
                resource = loaded_db.patient(patient_id)
                if resource:
                    observation['patient'] = resource['identifier'][0]['value']
                    for k, v in resource.items():
                        if k in ['id', 'subject', 'resourceType', 'identifier']:
                            continue
                        observation[f"patient_{k}"] = v

            if observation.get('focus'):
                focus = observation['focus']
                for f in focus:
                    if f['reference'].startswith('Procedure/'):
                        p = self.flattened_procedure(f['reference'])
                        for k, v in p.items():
                            if k in ['id', 'subject', 'resourceType']:
                                continue
                            observation[f"procedure_{k}"] = v
                        if "procedure_reason" in observation:
                            c = self.flattened_condition(observation["procedure_reason"])
                            for k, v in c.items():
                                if k in ['id', 'subject', 'resourceType']:
                                    continue
                                observation[f"condition_{k}"] = v
                            del observation["procedure_reason"]
                    if f['reference'].startswith('Specimen/'):
                        p = self.flattened_specimen(f['reference'])
                        for k, v in p.items():
                            if k in ['id', 'subject', 'resourceType']:
                                continue
                            observation[f"specimen_{k}"] = v

                del observation['focus']

            for coding_normalized, coding_source in normalize_coding(observation):
                observation[coding_source] = coding_normalized

            # renormalize the value in components
            if observation.get('component', []):
                for component in observation.get('component', []):
                    # simplify the value
                    observation_ = deepcopy(observation)
                    codings = normalize_coding(component)
                    for _ in codings:
                        coding_normalized, coding_source = _
                        if coding_source == 'code':
                            observation['code'] = coding_normalized
                            break
                    value_normalized, value_source = normalize_value(component)

                    observation_['value_numeric'] = None
                    observation_['value_normalized'] = None

                    if not value_normalized:
                        continue

                    value_numeric = value_normalized.split(' ')[0]

                    if value_source in observation_:
                        del observation_[value_source]

                    # remove any value[x] from parent observation
                    delete_value_x = [k for k in observation_ if k.startswith('value')]
                    for _ in delete_value_x:
                        del observation_[_]

                    observation_['value_normalized'] = value_normalized

                    if is_number(value_numeric):
                        observation_['value_numeric'] = float(value_numeric)
                    else:
                        observation_['value_numeric'] = None

                    if 'component' in observation_:
                        del observation_['component']
                    observation_['parent_observation'] = observation_['id']
                    observation_['id'] = uuid.uuid3(uuid.NAMESPACE_DNS, observation_['id'] + '/' + json.dumps(component, separators=(',', ':'))).hex

                    yield observation_
            else:
                if 'component' in observation:
                    del observation['component']
                else:
                    yield observation

        connection.close()

    def flattened_document_references(self) -> Generator[dict, None, None]:
        loaded_db = self
        connection = sqlite3.connect(loaded_db.db_name)
        cursor = connection.cursor()
        cursor.execute("SELECT * FROM resources where resource_type = ?", ("DocumentReference",))

        for _ in cursor.fetchall():
            key, resource_type, procedure = _
            document_reference = json.loads(procedure)

            # simplify the subject
            subject = document_reference['subject']['reference']
            document_reference['subject'] = subject

            # simplify the identifier
            document_reference['identifier'] = document_reference.get('identifier', [{'value': None}])[0]['value']

            # simplify the extensions
            for _ in document_reference['content'][0]['attachment']['extension']:
                value_normalized, value_source = normalize_value(_)
                document_reference[_['url'].split('/')[-1]] = value_normalized

            for k, v in document_reference['content'][0]['attachment'].items():
                if k in ['extension']:
                    continue
                document_reference[k] = v

            if subject.startswith('Patient/'):
                _, patient_id = subject.split('/')
                resources = [_ for _ in loaded_db.patient_everything(patient_id)]
                resources.append(loaded_db.patient(patient_id))
                for resource in resources:

                    if resource['resourceType'] == 'Patient':
                        document_reference['patient'] = resource['identifier'][0]['value']
                        continue

                    if resource['resourceType'] == 'Condition' and f"Condition/{resource['id']}" == procedure['reason']:
                        document_reference['reason'] = resource['code']['text']
                        continue

                    if resource['resourceType'] == 'Observation':
                        # must be focus
                        if f"Procedure/{procedure['id']}" not in [_['reference'] for _ in resource['focus']]:
                            continue

                        # TODO - pick first coding, h2 allow user to specify preferred coding
                        code = resource['code']['coding'][0]['code']

                        value = normalize_value(resource)

                        assert value is not None, f"no value for {resource['id']}"
                        document_reference[code] = value

                        continue

                    # skip these
                    if resource['resourceType'] in ['Specimen', 'Procedure', 'ResearchSubject', 'DocumentReference']:
                        continue

                    # default, add entire resource as an item of the list
                    resource_type = inflection.underscore(resource['resourceType'])
                    if resource_type not in procedure:
                        document_reference[resource_type] = []
                    document_reference[resource_type].append(resource)

            del document_reference['content']
            yield document_reference

        connection.close()


def create_dataframe(directory_path: str, work_path: str, data_type:str) -> pd.DataFrame:
    """Create a dataframe from the FHIR data in the directory."""
    assert pathlib.Path(work_path).exists(), f"Directory {work_path} does not exist."
    work_path = pathlib.Path(work_path)
    db_path = (work_path / "local_fhir.db")
    db_path.unlink(missing_ok=True)

    db = LocalFHIRDatabase(db_name=db_path)
    db.load_ndjson_from_dir(path=directory_path)

    if data_type == "DocumentReference":
        df = pd.DataFrame(db.flattened_document_references())
    elif data_type == "Observation":
        df = pd.DataFrame(db.flattened_observations())
    else:
        raise ValueError(f"{data_type} not supported yet. Supported data types are DocumentReference and Observation")
    assert not df.empty, "Dataframe is empty, are there any DocumentReference resources?"

    front_column_names = ["resourceType", "identifier"]
    if "patient" in df.columns:
        front_column_names = front_column_names + ["patient"]

    remaining_columns = [col for col in df.columns if col not in front_column_names]
    rear_column_names = ["status", "id"]
    if "subject" in df.columns:
        rear_column_names = rear_column_names + ["subject"]
    for c in df.columns:
        if c.endswith("_identifier"):
            rear_column_names.append(c)
    remaining_columns = [col for col in remaining_columns if col not in rear_column_names]

    reordered_columns = front_column_names + remaining_columns + rear_column_names
    df = df[reordered_columns]
    df = df.replace({np.nan: ''})
    return df


def normalize_value(resource_dict: dict) -> tuple[Optional[str], Optional[str]]:
    """return a tuple containing the normalized value and the name of the field it was derived from"""
    value_normalized = None
    value_source = None

    if set(resource_dict.keys()) == set(['url', 'extension']):
        assert len(resource_dict['extension']) > 0, f"Expected at least on extension, in {resource_dict}"
        return normalize_value(resource_dict['extension'][0])
    if 'valueQuantity' in resource_dict:
        value = resource_dict['valueQuantity']
        value_normalized = f"{value['value']} {value.get('unit', '')}"
        value_source = 'valueQuantity'
    elif 'valueCodeableConcept' in resource_dict:
        value = resource_dict['valueCodeableConcept']
        value_normalized = ' '.join([coding['display'] for coding in value.get('coding', [])])
        value_source = 'valueCodeableConcept'
    elif 'valueCoding' in resource_dict:
        value = resource_dict['valueCoding']
        value_normalized = value['display']
        value_source = 'valueCoding'
    elif 'valueString' in resource_dict:
        value_normalized = resource_dict['valueString']
        value_source = 'valueString'
    elif 'valueCode' in resource_dict:
        value_normalized = resource_dict['valueCode']
        value_source = 'valueCode'
    elif 'valueBoolean' in resource_dict:
        value_normalized = str(resource_dict['valueBoolean'])
        value_source = 'valueBoolean'
    elif 'valueInteger' in resource_dict:
        value_normalized = str(resource_dict['valueInteger'])
        value_source = 'valueInteger'
    elif 'valueRange' in resource_dict:
        value = resource_dict['valueRange']
        low = value['low']
        high = value['high']
        value_normalized = f"{low['value']} - {high['value']} {low.get('unit', '')}"
        value_source = 'valueRange'
    elif 'valueRatio' in resource_dict:
        value = resource_dict['valueRatio']
        numerator = value['numerator']
        denominator = value['denominator']
        value_normalized = f"{numerator['value']} {numerator.get('unit', '')}/{denominator['value']} {denominator.get('unit', '')}"
        value_source = 'valueRatio'
    elif 'valueSampledData' in resource_dict:
        value = resource_dict['valueSampledData']
        value_normalized = value['data']
        value_source = 'valueSampledData'
    elif 'valueTime' in resource_dict:
        value_normalized = resource_dict['valueTime']
        value_source = 'valueTime'
    elif 'valueDateTime' in resource_dict:
        value_normalized = resource_dict['valueDateTime']
        value_source = 'valueDateTime'
    elif 'valuePeriod' in resource_dict:
        value = resource_dict['valuePeriod']
        value_normalized = f"{value['start']} to {value['end']}"
        value_source = 'valuePeriod'
    # for debugging...
    # else:
    #     raise ValueError(f"value[x] not found in {resource_dict}")

    return value_normalized, value_source


def normalize_coding(resource_dict: dict) -> List[Tuple[str, str]]:
    def extract_coding(coding_list):
        return ' '.join([coding.get('display', '') for coding in coding_list if 'display' in coding])

    def find_codings_in_dict(d: dict, parent_key: str = '') -> List[Tuple[str, str]]:
        codings = []
        for key, value in d.items():
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        # Check if the dict contains a 'coding' list
                        if 'coding' in item and isinstance(item['coding'], list):
                            coding_string = extract_coding(item['coding'])
                            codings.append((coding_string, key))
                        # Recursively search in the dict
                        codings.extend(find_codings_in_dict(item, key))
            elif isinstance(value, dict):
                # Check if the dict contains a 'coding' list
                if 'coding' in value and isinstance(value['coding'], list):
                    coding_string = extract_coding(value['coding'])
                    codings.append((coding_string, key))
                # Recursively search in the dict
                codings.extend(find_codings_in_dict(value, key))
        return codings

    return find_codings_in_dict(resource_dict)

def is_number(s):
    """ Returns True if string is a number. """
    try:
        complex(s)
        return True
    except ValueError:
        return False



import pytest
from pydantic import ValidationError

from gen3_tracker.common import ACCEPTABLE_HASHES
from gen3_tracker.git import DVCItem

VALID_HASHES = {
    'md5': 'acbd18db4cc2f85cedef654fccc4a4d8',
    'sha1': '2ef7bde608ce5404e97d5f042f95f89f1c232871',
    'sha256': '5bf8aa57fc5a6bc547decf1cc6db63f10deb55a3c6c5df497d631fb3d95e1abf',
    'sha512': '3ba2942ed1d05551d4360a2a7bb6298c2359061dc07b368949bd3fb7feca3344221257672d772ce456075b7cfa50fd7ce41eaefe529d056bf23dd665de668b78',
    'crc': '3e25960a',
    'etag': 'acbd18db4cc2f85cedef654fccc4a4d8-3'
}


def test_invalid_hash_values():
    """Test that invalid hash values raise a ValidationError."""
    for hash_type in ACCEPTABLE_HASHES.keys():
        _ = dict(hash=hash_type, modified='2013-07-01T16:10-04:00', path='dddd', size=1)
        _[hash_type] = 'foo'
        print(_)
        with pytest.raises(ValidationError):
            item = DVCItem(**_)
            print(item)


def test_valid_hash_values():
    """Test that valid hash values do raise a ValidationError."""
    for hash_type in VALID_HASHES.keys():
        _ = dict(hash=hash_type, modified='2013-07-01T16:10-04:00', path='dddd', size=1)
        _[hash_type] = VALID_HASHES[hash_type]
        print(_)
        item = DVCItem(**_)
        print(item)



import asyncio
import json
import logging
import pathlib
from datetime import datetime
import time
from urllib.parse import urlparse
from zipfile import ZipFile

from gen3.auth import Gen3Auth
from gen3.jobs import Gen3Jobs

from gen3_tracker import Config
from gen3_tracker.common import Push, Commit
from gen3_tracker.gen3.indexd import write_indexd
from gen3_tracker.git import calculate_hash, DVC, run_command, DVCMeta, DVCItem, modified_date


def _validate_parameters(from_: str) -> pathlib.Path:

    assert len(urlparse(from_).scheme) == 0, f"{from_} appears to be an url. url to url cp not supported"

    return from_


def cp(config: Config,
       from_: str,
       project_id: str,
       ignore_state: bool,
       auth=None,
       user=None,
       object_name=None,
       bucket_name=None,
       metadata: dict = {}
       ):
    """Copy meta to bucket, used by etl_pod job"""
    from_ = _validate_parameters(str(from_))
    if not isinstance(from_, pathlib.Path):
        from_ = pathlib.Path(from_)

    assert auth, "auth is required"

    metadata = dict({'submitter': None, 'metadata_version': '0.0.1', 'is_metadata': True} | metadata)
    if not metadata['submitter']:
        if not user:
            user = auth.curl('/user/user').json()
        metadata['submitter'] = user['name']

    program, project = project_id.split('-')

    assert bucket_name, f"could not find bucket for {program}"

    temp_dir = config.work_dir
    temp_dir = pathlib.Path(temp_dir)

    if not object_name:
        now = datetime.now().strftime("%Y%m%d-%H%M%S")
        object_name = f'_{project_id}-{now}_meta.zip'

    zipfile_path = temp_dir / object_name
    with ZipFile(zipfile_path, 'w') as zip_object:
        for _ in from_.glob("*.ndjson"):
            zip_object.write(_)

    stat = zipfile_path.stat()
    md5_sum = calculate_hash('md5', zipfile_path)
    my_dvc = DVC(
            meta=DVCMeta(),
            outs=[
                DVCItem(
                    path=object_name,
                    md5=md5_sum,
                    hash='md5',
                    modified=modified_date(zipfile_path),
                    size=stat.st_size,

                )
            ]
        )

    metadata = write_indexd(
        auth=auth,
        project_id=project_id,
        bucket_name=bucket_name,
        overwrite=False,
        restricted_project_id=None,
        dvc=my_dvc,
    )

    # document = file_client.upload_file_to_guid(guid=id_, file_name=object_name, bucket=bucket_name)
    # print(document, file=sys.stderr)

    run_command(f"gen3-client upload-single --bucket {bucket_name} --guid {my_dvc.object_id} --file {zipfile_path} --profile {config.gen3.profile}", no_capture=False)

    return {'msg': f"Uploaded {zipfile_path} to {bucket_name}", "object_id": my_dvc.object_id, "object_name": object_name}


def publish_commits(config: Config, wait: bool, auth: Gen3Auth, bucket_name: str, spinner=None) -> dict:
    """Publish commits to the portal."""

    # TODO legacy fhir-import-export job: copies meta to bucket and triggers job,
    #  meta information is already in git REPO,
    #  we should consider changing the fhir_import_export job to use the git REPO

    user = auth.curl('/user/user').json()

    # copy meta to bucket
    upload_result = cp(
        config=config,
        from_='META',
        project_id=config.gen3.project_id,
        ignore_state=True,
        auth=auth,
        user=user,
        bucket_name=bucket_name
    )

    object_id = upload_result['object_id']

    push = Push(config=config)
    jobs_client = Gen3Jobs(auth_provider=auth)

    push.commits.append(Commit(object_id=object_id, message='From g3t-git', meta_path=upload_result['object_name'], commit_id=object_id))
    args = {'push': push.model_dump(), 'project_id': config.gen3.project_id, 'method': 'put'}

    from cdislogging import get_logger  # noqa
    cdis_logging = get_logger("__name__")
    cdis_logging.setLevel(logging.WARN)

    if wait:
        _ = asyncio.run(jobs_client.async_run_job_and_wait(job_name='fhir_import_export', job_input=args, spinner=spinner))
    else:
        _ = jobs_client.create_job('fhir_import_export', args)
    if not isinstance(_, dict):
        _ = {'output': _}
    if isinstance(_['output'], str):
        try:
            _['output'] = json.loads(_['output'])
        except json.JSONDecodeError:
            pass
    return _


# monkey patch for gen3.jobs.Gen3Jobs.async_run_job_and_wait
# make it less noisy and sleep less (max of 30 seconds)
async def async_run_job_and_wait(self, job_name, job_input, spinner=None, _ssl=None, **kwargs):
    """
    Asynchronous function to create a job, wait for output, and return. Will
    sleep in a linear delay until the job is done, starting with 1 second.

    Args:
        _ssl (None, optional): whether or not to use ssl
        job_name (str): name for the job, can use globals in this file
        job_input (Dict): dictionary of input for the job

    Returns:
        Dict: Response from the endpoint
    """
    job_create_response = await self.async_create_job(job_name, job_input)
    status = {"status": "Running", "name": job_name}
    initial_sleep_time = 3
    sleep_time = initial_sleep_time
    max_sleep_time = 30
    while status.get("status") == "Running":
        if spinner:
            spinner.text = f"{status.get('name')} waiting {int(sleep_time)}"
        else:
            logging.debug(f"job still running, waiting for {sleep_time} seconds...")
        time.sleep(sleep_time)
        sleep_time *= 1.5
        if sleep_time > max_sleep_time:
            sleep_time = initial_sleep_time
        status = await self.async_get_status(job_create_response.get("uid"))
        if not spinner:
            logging.info(f"{status}")

    if not spinner:
        logging.info(f"Job is finished!")
    else:
        spinner.text = f"{status.get('name')} {status.get('status')}"

    if status.get("status") != "Completed":
        raise Exception(f"Job status not complete: {status.get('status')}.")

    response = await self.async_get_output(job_create_response.get("uid"))
    return response


Gen3Jobs.async_run_job_and_wait = async_run_job_and_wait



import json
import os
import pathlib

from tests.integration import run
from click.testing import CliRunner
import pytest


SHOULD_SUCCEED = [
    "--debug add s3://s3-bucket/s3-object --size 1 --modified 2024-05-05T07:26:29-0700 --md5 acbd18db4cc2f85cedef654fccc4a4d8",
    "--debug add https://example.com/http-bucket/http-object --size 1 --modified 2024-05-05T07:26:29-0700 --sha1 2ef7bde608ce5404e97d5f042f95f89f1c232871",
    "--debug add azure://azure-bucket/azure-object --size 1 --modified 2024-05-05T07:26:29-0700 --sha512 3ba2942ed1d05551d4360a2a7bb6298c2359061dc07b368949bd3fb7feca3344221257672d772ce456075b7cfa50fd7ce41eaefe529d056bf23dd665de668b78",
    "--debug add gs://gs-bucket/gs-object --size 1 --modified 2024-05-05T07:26:29-0700 --etag acbd18db4cc2f85cedef654fccc4a4d8-3",
]

EXPECTED_MANIFEST_PATHS = [
    "MANIFEST/s3-bucket/s3-object.dvc",
    "MANIFEST/http-bucket/http-object.dvc",
    "MANIFEST/azure-bucket/azure-object.dvc",
    "MANIFEST/gs-bucket/gs-object.dvc",
]

SHOULD_FAIL = [
    # unsupported scheme
    "--debug add foo://s3-bucket/s3-object --size 1 --modified 2024-05-05T07:26:29-0700 --md5 acbd18db4cc2f85cedef654fccc4a4d8",
    # no size
    "--debug add s3://s3-bucket/s3-object  --modified 2024-05-05T07:26:29-0700 --md5 acbd18db4cc2f85cedef654fccc4a4d8",
    # negative size
    "--debug add https://example.com/http-bucket/http-object --size -1 --modified 2024-05-05T07:26:29-0700 --sha1 2ef7bde608ce5404e97d5f042f95f89f1c232871",
    # invalid modified
    "--debug add azure://azure-bucket/azure-object --size 1 --modified XXX --sha512 3ba2942ed1d05551d4360a2a7bb6298c2359061dc07b368949bd3fb7feca3344221257672d772ce456075b7cfa50fd7ce41eaefe529d056bf23dd665de668b78",
    # invalid hash
    "--debug add gs://gs-bucket/gs-object --size 1 --modified 2024-05-05T07:26:29-0700 --etag foo",
]


def test_bucket_import(runner: CliRunner, project_id, tmpdir) -> None:
    """Test import from sources other than filesystem."""
    # change to the temporary directory
    assert tmpdir.chdir()
    print(pathlib.Path.cwd())

    print(project_id)

    run(runner, ["--debug", "init", project_id, "--approve", "--no-server"],
        expected_files=[".g3t", ".git"])

    for _ in SHOULD_SUCCEED:
        run(runner, _.split())

    for _ in EXPECTED_MANIFEST_PATHS:
        _ = pathlib.Path(_)
        assert _.exists(), f"{_} does not exist."

    for _ in SHOULD_FAIL:
        with pytest.raises(Exception):
            run(runner, _.split())

    # test the ls command
    result = run(runner, ["--debug", "--format", "json",  "ls"])
    listing = json.loads(result.stdout)

    for _ in ['bucket', 'committed', 'uncommitted']:
        assert _ in listing

    # files should appear in uncommitted
    assert len(listing['uncommitted']) == len(SHOULD_SUCCEED)

    # commit the changes
    run(runner, ["--debug", "commit", "-am", "initial commit"])

    # test the ls command, should now be in committed
    result = run(runner, ["--debug", "--format", "json",  "ls"])
    listing = json.loads(result.stdout)
    assert len(listing['committed']) == len(SHOULD_SUCCEED)

    # test the ls filter
    for _ in EXPECTED_MANIFEST_PATHS:
        bucket_name = _.split('/')[1]
        result = run(runner, ["--debug", "--format", "json",  "ls", bucket_name])
        listing = json.loads(result.stdout)
        assert len(listing['committed']) == 1



from gen3_tracker.git import to_dvc, DVCItem


def test_read_dvc():
    dvc = to_dvc('tests/fixtures/hello.txt.dvc')
    assert dvc
    assert dvc.outs
    assert dvc.outs[0].path == 'my-project-data/hello.txt'


def test_read_dvc_item():
    _ = {'hash': 'md5', 'is_symlink': False, 'md5': 'b1946ac92492d2347c6235b4d2611184', 'mime': 'text/plain', 'modified': '2024-04-30T17:46:30.819143+00:00', 'path': 'my-project-data/hello.txt', 'realpath': '/Users/walsbr/aced/g3t-git/attic/cbds-test39/my-project-data/hello.txt', 'size': 6}
    item = DVCItem(**_)
    assert item
    assert item.hash == 'md5'



import os
import pathlib

import yaml
from click.testing import CliRunner

from gen3_tracker.config import ensure_auth, default
from gen3_tracker.git import DVC, run_command
from tests.integration import run, validate_document_in_psql_graph, validate_document_in_elastic


def test_simple_workflow(runner: CliRunner, project_id, tmpdir) -> None:
    """Test the init command."""
    # change to the temporary directory
    assert tmpdir.chdir()
    print(pathlib.Path.cwd())

    assert os.environ.get("G3T_PROFILE"), "G3T_PROFILE environment variable must be set."

    print(project_id)

    run(runner, ["--debug", "init", project_id, "--approve"],
        expected_files=[".g3t", ".git"])

    # create a test file
    test_file = pathlib.Path("my-project-data/hello.txt")
    test_file.parent.mkdir(parents=True, exist_ok=True)
    test_file.write_text('hello\n')

    # add the file
    run(runner, ["--debug", "add", str(test_file)], expected_files=["MANIFEST/my-project-data/hello.txt.dvc"])

    # should create a dvc file
    dvc_path = pathlib.Path("MANIFEST/my-project-data/hello.txt.dvc")
    assert dvc_path.exists(), f"{dvc_path} does not exist."
    with open(dvc_path) as f:
        yaml_data = yaml.safe_load(f)
    assert yaml_data
    dvc = DVC.model_validate(yaml_data)
    assert dvc, "DVC file not parsed."

    # capture expected object_id
    dvc.project_id = project_id
    object_id = dvc.object_id

    # create the meta file
    run(runner, ["--debug", "meta", "init"], expected_files=["META/DocumentReference.ndjson"])

    # commit the changes, delegating to git
    run(runner, ["--debug", "commit", "-am", "initial commit"])

    # validate the meta files
    run(runner, ["--debug", "meta", "validate"])

    # create a visualisation
    run(runner, ["--debug", "meta", "graph"], expected_files=["meta.html"])

    # create a dataframe
    run(runner,  ["--debug", "meta", "dataframe", '--data_type', 'DocumentReference'], expected_files=["meta.csv"])

    # push to the server
    run(runner, ["--debug", "push"])

    # list the files from indexd
    run(runner, ["--debug", "ls"], expected_output=["my-project-data/hello.txt"])

    # check the files exist in the graph and flat databases
    auth = ensure_auth(config=default())
    validate_document_in_psql_graph(object_id, auth=auth)
    validate_document_in_elastic(object_id, auth=auth)

    # clone the project

    # create a new directory, cd to it
    os.mkdir("clone")
    os.chdir("clone")

    # clone the project
    run(runner, ["--debug", "clone", project_id])
    # pull the data
    run(runner, ["--debug", "pull"])
    # check the files exist in the cloned directory
    run_command("ls -l")
    assert pathlib.Path("my-project-data/hello.txt").exists(), "hello.txt does not exist in the cloned directory."

    # remove the project from the server.
    # TODO note, this does not remove the files from the bucket (UChicago bug)
    # See https://ohsucomputationalbio.slack.com/archives/C043HPV0VMY/p1714065633867229
    run(runner, ["--debug", "projects", "empty", "--project_id", project_id, "--confirm", "empty"])
    run(runner, ["--debug", "projects", "rm", "--project_id", project_id])

    # TODO fix `collaborator rm`
    # arborist logs:  "Policy `data_upload` does not exist for user `xxx@xxx.xxx`: not revoking. Check if it is assigned through a group."
    # username = auth.curl('/user/user').json()['username']
    # run(runner, ["--debug", "collaborator", "rm", username, "--approve"], expected_output=[username])

    # add a user with write permissions
    run(runner, ["--debug", "collaborator", "add", "foo@bar.com", "--write", "--approve"])

    # add a user from another directory (without config)
    os.mkdir("empty")
    os.chdir("empty")
    program, project = project_id.split("-")
    run(runner, ["--debug", "collaborator", "add", "foo2@bar.com", f"/programs/{program}/projects/{project}", "--write", "--approve"])



import pathlib
from click.testing import CliRunner, Result

from gen3_tracker.cli import cli
from gen3_tracker.config import ensure_auth, default
from gen3.submission import Gen3Submission
from gen3.query import Gen3Query


def run(runner: CliRunner, args: list[str], expected_output: list[str] = [], expected_exit_code: int = 0, expected_files: list[pathlib.Path] = []) -> Result:
    """Run a command and check the output, exit code and expected files."""
    if isinstance(args, str):
        args = args.split()
    if isinstance(expected_output, str):
        expected_output = expected_output.splitlines()
    if isinstance(expected_files, pathlib.Path):
        expected_files = [expected_files]
    expected_files = [pathlib.Path(_) for _ in expected_files]

    print('------------------------------------------------------------')
    print("g3t " + " ".join(args))
    result = runner.invoke(cli, args)
    print(result.stdout)
    assert result.exit_code == expected_exit_code, f"exit_code: {result.exit_code}, expected: {expected_exit_code}"
    for line in expected_output:
        assert line in result.output, f"output: {result.output}, expected: {expected_output}"
        print(f"{line} found in output.")
    for file in expected_files:
        assert file.exists(), f"{file} does not exist."
        print(f"{file} exists.")

    return result


def validate_document_in_psql_graph(did: str, auth=None):
    """Simple query to validate a document in the graph."""
    if not auth:
        auth = ensure_auth(config=default())
    gen3_submission = Gen3Submission(auth)
    result = gen3_submission.query(query_txt="""
        {
            document_reference(id:"DID") {
                id
            }
        }
    """.replace("DID", did))
    print(result)
    assert result['data']['document_reference'][0]['id'] == did


def validate_document_in_elastic(did, auth):
    """Simple query to validate a document in elastic."""
    query = Gen3Query(auth)
    result = query.graphql_query(
        query_string="""
            query($filter:JSON) {
              file(filter:$filter) {
                id
              }
            }
        """,
        variables={"filter": {"AND": [{"IN": {"id": [did]}}]}}
    )
    print(result)
    assert result['data']['file'][0]['id'] == did
